 /*
  * Software for OBI Plug with WiFi ESP Chip
  * 
  * by TomT
  * contact : info@tomz.ws
  * 
  * First BETA Revision
  */
  
  /*
  * Default SoftAP IP is : 192.168.4.1
  * Default SoftAP Port is : 80
  * Default SoftAP Password is : esp8266-1
  * 
  * COM Port : 115200 Baud 8N1
  */

/* 
 *  ########## Pins ############
 *  GPIO4 : LED
 *  GPIO5 : Relay OFF
 *  GPIO12 : Relay ON
 *  GPIO14 : Push Button
 */


   /*
    * ################ READ ME BEFORE ######################
    * 
    *  Please use the foloowing settings to flash the program to your OBI Plug:
    *  * ##### programming settings for ARDUINO #####
    *  Board : generic ESP8266 Module
    *  Flash Mode : DIO
    *  Flash Size : 1M (64k SPIFFS)
    *  Debug Port : disabled
    *  Debug Level : none
    *  Reset Method : ck
    *  Crystal frequency : 26 MHz 
    *  Flash frequency : 40 MHz 
    *  CPU frequency : 80 MHz 
    *  Upload Speed : 115200
    *  
    *  After flash is done, you can remove the Plug from your computer USB UART Connection.
    *  Plug into a mains power wall plug.
    *  
    *  Look for a Wifi Access point calles somewhat like ESP_XYZ
    *  Make a connection to this Access Point and type into your browser
    *  192.168.4.1.
    *  You will see a page where you can choose your Homenetwork.
    *  Type in the Name and SSID Password and after that reboot the ESP.
    *  
    *  Then it will login into your local network and you now need to find the correct IP Adress.
    *  Type in the IP Adress and a page will be shown, wehere you can configure and Switch ON / OFF the Obi Plug.
    *  
    *  In future i will add functions.
    *  The OTA Over the air Update is integrated and you can updtae to a new Firmware version without opening the Plug again. 
    *  
    *  DISCLAIMER / WARRANTY:
    *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
    *  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, 
    *  TITLE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING 
    *  THE SOFTWARE BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, 
    *  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE 
    *  OR OTHER DEALINGS IN THE SOFTWARE.
    *  
    *  USE ANYTHING HERE AT YOUR OWN RISK....
    *  
    * 
    * #########################################################
   */

  /*
    * #########################################################
    * 
    * TO BE IMPLEMENTED :
    * 
    * - ON Send Command
    * - OFF Send Command
    * 
  */
  
  /*
    * #########################################################
    * 
    * VERSION HISTORY :
    * 
    * 1.0.4
    * - added to try to reconnect to a network after 3 Minutes running AP Point (fix for slow WLAN Startup after MAINS Power lost of ROUTER)
    * - CCU CUXD Implementation
    * 
  */

/*
 * DON'T CHANGE ANYTHING BELOW!
 * EVERYTHING WILL BE CONBFIGURED LATER WITH YOUR WEBBROWSER 
*/


#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <WiFiClient.h>
#include <ESP8266mDNS.h>
#include <ESP8266HTTPUpdateServerTom.h>



   
//System  **** DONT CHANGE THIS or system will fail to run ****
const String VERSION = "BETA 1.0.4";         //Thats the actual Version of FW
String wwwInfos = "";                        //will be filled in later with actual data

//String S ="";
String wwwStatus = "";                       //takes the last Status message
String wwwStatus1 = "";                       //takes the last Status message
String wwwStatus2 = "";                       //takes the last Status message
String wwwStatus3 = "";                       //takes the last Status message
String wwwStatus4 = "";                       //takes the last Status message
String wwwCommandHeader = "";
int first_Startup = 0;                       //Do we start at the first time ever? 

String wwwGoBack = "";                       //holds the code for the back to root button 
int eeprom_Function = 0;                     //define what kind of data to be written to the EEPROM
int stateONPowerON = 0;                      //define the Realy state on Power ON 0=OFF 1= ON 

unsigned long dSendState = 600 * 1000;      //define the time for StatusMessage to ioBroker / msec.
unsigned long lSendState = 0;                //timestamp of last StatusMessage / msec.
unsigned long timeNow = 0;                   //timestamp for timeroperations / msec.

unsigned long dKeepWIFIAlive = 900 * 1000;   //Reconnect Delta Time / msec. (default 15 min.)
unsigned long lKeepWIFIAlive = 0;            //timestamp of last Reconnect / msec.
int enableReconnect = 1;                     //switch Wifi.reconnect function on or off 0=OFF 1=ON

unsigned long dTimeButton = 250;             //Button Check Delta Time/ msec.
unsigned long lTimeButton = 0;               //timestamp of last Button Check / msec.

String relStatus = "OFF";                    //holds the Relay Status 
String dStatus ="";                         
boolean ButtonIsBlocked = false;             //We dont want a blinking switch when button wasn't released

unsigned long iDelayON = 0;
unsigned long iDelayOFF = 0;
unsigned long iONDelayOFF = 0;
unsigned long iOFFDelayON = 0;
unsigned long iAutoONOFF = 0;
unsigned long dAutoONOFF = 0;
boolean enableDelay = false;
int handle_Delay = 0;
 
//ioBroker
char host_ioBrokerIP[16] = "192.168.4.1";     //IP Adress ioBroker
int host_ioBrokerPORT = 8087;                 //Portadress simple API Adapter
char IOB_Datapoint_Path[80] = "esp8266.0";    //Name of ioBroker Datapoint path
char IOB_Datapoint_Sub[80] = "Sub";          //Name of ioBroker Subfolder
char IOB_Datapoint_State[80] = "State";       //Name of ioBroker Status datapoint
char IOB_Datapoint_Message[80] = "Info";      //Name of ioBroker Message datapoint
String IOB_State ="";                         //ioBroker State Informations
String IOB_Message ="";                       //ioBroker String to be sent to Message Datapoint 
int IOB_active = 0;                           //is ioBroker State Report active? 0=no 1=yes

//CCU 
int CCU_CuxD = 0;                              //Send state to CuxD enabled 0=disabled 1=enabled
char CCU_CuxDAdress[38] = "CUX2801001:1";      //Name of CCU CuxD Device Adress incl. Channel
char host_CCUIP[16] = "192.168.4.1";               //IP Adress CCU2

ESP8266WebServer server(80);
ESP8266HTTPUpdateServer httpUpdater;


// We create a Timestamp
String timeStamp() { 
      char stamp[10];
      unsigned long tHour = millis()/3600000;
      unsigned long tMinute = millis()/60000-tHour*60;
      unsigned long  tSecond = millis()/1000-tHour*3600-tMinute*60;
      sprintf (stamp,"%03d:%02d:%02d", tHour, tMinute, tSecond);
      return stamp;
}



String wwwHeader = "<!DOCTYPE HTML>\n"
                  "<html><head><meta content=\"text/html;charset=utf-8\"><title>ESP8266 Wall Plug OBI</title>\n"
                  "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">"
                  "<style type=\"text/css\">body {text-align:center;font-family: sans-serif;background-color: #000000;color: #AAAAAA;font-size:0.9em;} a:link, a:visited, a:hover, a:active {color:#fffff;}</style>"
                  "</head>\n"
                  "<body>\n"
                  "<center>\n"
                  "<table border=1 width=\"80%\" cellpadding=\"6\"><tr><td colspan=2>"
                  "<h2>ESP8266 Wall Plug OBI</h2>"
                  "</td></tr>\n\n";


String wwwFooter =  "</table><br><hr>\n"
                   "<p style=\"font-size:0.6em\" a:link, a:visited, a:hover, a:active {color:#fffff;>2018 by TomT - - - Version " + VERSION + " - - - contact me : <a href=\"http://www.tomz.ws\" target=_new>http://www.tomz.ws</a> or send email to : <a href=\"mailto:info@tomz.ws\">info@tomz.ws</a><br>"
                   "AP for WIFI settings is found at 192.168.4.1, if the ESP is not connected to a network. "
                    "If ESP fails to connect to a network, it will start an AP after 60 sec. timeout<br></p></center>\n"
                   "</body></html>";




String wwwCommandlist = "<tr><td colspan=2><p style=\"font-size:0.8em\"><b>Commands for setup : &nbsp&nbsp&nbsp&nbsp&nbsp</b>\n"
                        "<b><a href=/reboot title=\"restart the ESP\">[REBOOT]</a> : </b>\n"
                        "<b><a href=/erasessid title=\"clear the SSID and PASSWORD, reboot and open an AP for config\">[ERASE SSID]</a> : </b>\n"
                        "<b><a href=/eraseEEPROM title=\"clear the whole EEPROM, reboot and open an AP for config\">[ERASE EEPROM]</a> : </b> \n"
                        "<b><a href=/update title=\"Update Firmware OverTheAir\">[FIRMWARE UPDATE]</a></b>\n"
                        "</td></tr><tr><td valign=top><p style=\"font-size:0.8em\">"
                        "<b>/ON</a> : </b>Switch permanently ON<br><b>/OFF</a> : </b>Switch permanently OFF<br>"
                        "<b>/delayON?=&lt;value&gt;</b> set the Relay delayed ON (in sec.)<br>\n"
                        "<b>/delayOFF?=&lt;value&gt;</b> set the Relay delayed OFF (in sec.)<br>\n"
                        "<b>/ONdelayOFF?=&lt;value&gt;</b> set the Relay instant ON then delayed OFF (in sec.)<br>\n"
                        "<b>/OFFdelayON?=&lt;value&gt;</b> set the Relay instant OFF then delayed ON (in sec.)<br>\n"
                         "<b>/AutoONOFF?=&lt;value&gt;</b> set the Relay repeated ON-OFF mode (in sec.)<br>\n"
                        "<b>/keepalive?=&lt;value&gt;</b> Set Wifi.reconnect time in sec.<br>\n"
                        "<b>/reConnect?=</b> Wifi:reconnect Enabled 0=OFF 1=ON<br>\n"
                        "<b>/powerONState?=&lt;value&gt;</b> set the Relay PowerON State (0=OFF 1=ON)</p>\n"
                        "</td><td valign=top><p style=\"font-size:0.8em\">\n"
                        "<b>/ioBroker?active=&lt;value&gt;</b> send state ioBroker? 0=no 1=yes<br>\n"
                        "<b>/ioBroker?ip=&lt;value&gt;</b> set the ioBroker IP Adress<br>\n"
                        "<b>/ioBroker?port=&lt;value&gt;</b> set the ioBroker SimpleAPI Port<br>\n"
                        "<b>/ioBroker?path=&lt;value&gt;</b> ioBroker Datapoint Name\n<br>"
                        "<b>/ioBroker?sub=&lt;value&gt;</b> ioBroker Datapoint Subpath\n<br>"
                        "<b>/ioBroker?st=&lt;value&gt;</b> ioBroker Datapoint State Name\n<br>"
                        "<b>/ioBroker?msg=&lt;value&gt;</b> ioBroker Datapoint Message Name\n<br>"
                        "<b>/ioBroker?statetime=&lt;value&gt;</b> ioBroker StateReport Time (min. 60 sec.)\n<br><br>"
                        "<b>/CCU?ip=&lt;value&gt;</b> set the CCU IP Adress\n<br>"
                        "<b>/CCU?cuxdDevice=&lt;value&gt;</b> set the CUXD Devise with channel number\n<br>"
                        "<b>/CCU?cuxd=&lt;value&gt;</b>send state to CuxD? 0=NO 1=YES\n<br><br>"
                        "</p></td></tr>\n";

                        

void handle_root() {
  wwwInfos = "<tr><td><b>ESP8266</b><br><br> own IP : " + WiFi.localIP().toString() + "<br>";
  wwwInfos = wwwInfos + "GATEWAY : " + WiFi.SSID() +"<br> IP : " + WiFi.gatewayIP().toString()  + " <br>RSSI : " + WiFi.RSSI() + " dBm<br>FlashSize : " + ESP.getFlashChipSize() + "<br><br>Time since start : " + timeStamp() + "<br><br>";
  wwwInfos = wwwInfos + "<p style=\"font-size:1.5em\"><b>SWITCH STATUS : "+ relStatus + "</b> " + dStatus + "</p>";
  wwwInfos = wwwInfos + "<b><a href=/ON>SWITCH ON</a><br>";
  wwwInfos = wwwInfos + "<a href=/OFF>SWITCH OFF</a></b><br>";
  wwwInfos = wwwInfos + "</td>";
  wwwInfos = wwwInfos + "<td><p style=\"font-size:0.8em\"><b>SETTINGS :</b></p>";
  wwwInfos = wwwInfos + "PowerONState : " + stateONPowerON + "<br>" + "Wifi-Reconnect enabled : " + enableReconnect + "<br>" + "Wifi.Reconnect time : " + dKeepWIFIAlive / 1000 + " sec.<br><br>";
  wwwInfos = wwwInfos + "<b>ioBroker : </b><br>"  + "<br>" + "Active : " + String(IOB_active) + "<br>" + "IP : " + String(host_ioBrokerIP) + "<br>" +" SimpleAPI Port : " + host_ioBrokerPORT + "<br>";
  wwwInfos = wwwInfos + "Datapoint Name : " + IOB_Datapoint_Path + "<br>" + "Subpath : " + IOB_Datapoint_Sub + "<br>" + "State Name : " + IOB_Datapoint_State + "<br>" + "Message Name : " + IOB_Datapoint_Message + "<br>";
  wwwInfos = wwwInfos + "State Time Report : " + (dSendState / 1000) + " sec.<br><br>";
  wwwInfos = wwwInfos + "<b>CCU : </b><br>"  + "<br>" + "IP : " + String(host_CCUIP) + "<br>" + " CuxD Device : " + String(CCU_CuxDAdress) + "<br>" + " CuxD enabled : " + CCU_CuxD + "<br>";
  wwwInfos = wwwInfos + "</td></tr>";
  wwwInfos = wwwInfos + "<tr><td colspan=2><p style=\"font-size:0.8em\">Last States : " + wwwStatus + "</p></td></tr>";
  server.send(200, "text/html", wwwHeader + wwwInfos + wwwCommandlist + wwwFooter);
  delay(100);
}



void setup() {

  Serial.begin (115200);

  WiFi.persistent(false);  // set WiFi credentials persistance to false to prevent eeprom corruption
  
  WiFi.setOutputPower(20.5);

  WiFi.mode(WIFI_STA);
  espNKWiFiconnect();
  
  char host[32];
  String tmpValue = WiFi.SSID();
  tmpValue.toCharArray(host, 32);
  
  MDNS.begin(host);

  Serial.println("IP address: "+ WiFi.localIP().toString());
  //Serial.println(WiFi.localIP());

  server.on("/", handle_root);
  server.on("/reboot", handle_APrestart);
  server.on("/erasessid", handle_clearAPeeprom);
  server.on("/eraseEEPROM", handle_clearAPalleeprom);
  server.on("/powerONState", handle_PowerONState);
  server.on("/reConnect", handle_Reconnect);
  server.on("/keepalive", handle_KeepAlive);
  server.on("/ioBroker", handle_ioBroker);   //ioBroker Handle
  server.on("/ON", switch_ON);               //ON
  server.on("/OFF", switch_OFF);             //OFF
  server.on("/delayON", delay_ON);           //delayed ON 
  server.on("/delayOFF", delay_OFF);         //delayed OFF
  server.on("/ONdelayOFF", ON_delay_OFF);    //direct ON and delayed OFF  
  server.on("/OFFdelayON", OFF_delay_ON);    //direct OFF and delayed ON
  server.on("/AutoONOFF", AUTO_ON_OFF);      //repeated AUTO ON/OFF
  server.on("/CCU", handle_CCU);             //CCU Handle

  
  //Send command ON
  server.on("/CommandON", command_ON);
  //Send command OFF
  server.on("/CommandOFF", command_OFF);


  httpUpdater.setup(&server);
  server.begin();
  Serial.println("HTTP server started");
  //wwwStatus = timeStamp() + " HTTP server started";
  build_wwwStatus("HTTP server started");
  
  MDNS.addService("http", "tcp", 80);

  readFirstStart();
  delay(100);
  firstStart();
  
  timeNow = millis();
  lKeepWIFIAlive = timeNow;
  lTimeButton = timeNow;
 
  //set the Pins to input or output mode
  pinMode(14, INPUT);  
  pinMode(4, OUTPUT);  
  pinMode(5, OUTPUT);  
  pinMode(12, OUTPUT); 
  digitalWrite(5, 1); 
  digitalWrite(12, 1);

  readEEPROM();  
  setInitialSwitchState();
  URL_GoBack();
  createCommandHeader();
}


void loop() {
  
  timeNow = millis();
  server.handleClient();
  delay(10);


   // Should that be a fix of Connection fails to the webserver?
   // we give it a try
   // see https://github.com/esp8266/Arduino/issues/2330
   //or use ping script that pings the ESP every 30 sec. (CuxD Ping device) 
   // we test with 2 minutes
   // 30sec Test was ok
   if (enableReconnect== 1) {
     if (timeNow - lKeepWIFIAlive > dKeepWIFIAlive) {
       WiFi.reconnect();
       lKeepWIFIAlive = timeNow;
       delay(10);
       
       build_wwwStatus ("Wifi.Reconnect done");
     }
   }
   
   // Check for a button press once a sec. 
   if (timeNow - lTimeButton > dTimeButton) {
     checkButton();
     lTimeButton = timeNow;
   }
   
   
   if (enableDelay == true){
    // Check if there was a DelayON command 
    if (timeNow > iDelayON && handle_Delay == 1) {delayswitch_ON();enableDelay = false;}
    // Check if there was a DelayOFF command 
    if (timeNow > iDelayOFF && handle_Delay == 2) {delayswitch_OFF();enableDelay = false;}
    // Check if there was a ONDelayOFF command 
    if (timeNow > iONDelayOFF && handle_Delay == 3) {delayswitch_OFF();enableDelay = false;}
    // Check if there was a OFFDelayON command 
    if (timeNow > iOFFDelayON && handle_Delay == 4) {delayswitch_ON();enableDelay = false;}
    // Check if there was a AutoONOFF command 
    if (timeNow > iAutoONOFF && handle_Delay == 5) {
      iAutoONOFF = iAutoONOFF  + dAutoONOFF;      
      if (relStatus == "ON"){delayswitch_OFF();}
      else {delayswitch_ON();}
    }
   }
   
   // timeLoop for StatusReports to IOB
   if (timeNow - lSendState > dSendState) {
     reportIOB();
     lSendState = timeNow;
     delay(10);
     build_wwwStatus ("sent Status to ioBroker");
   }
   

}

//check for a button press 
void checkButton(){

  int Button = 1;

  //debounce loop to prevent false positive pulses 
  int ButtonPress = 0;
  for (int i=0; i<16; i++){
    Button = digitalRead(14);
    if (Button == 0){ButtonPress++;}
  }
  //to prevent ON/OFF loop we switch once and wait for the button to be released
  if (ButtonPress >= 15 & ButtonIsBlocked == false){
    if (relStatus == "OFF"){
      //Serial.println("Button ON");
      switch_ON();
      ButtonIsBlocked = true;
    }
    else {
    //Serial.println("Button OFF");
    switch_OFF();
    ButtonIsBlocked = true;
    }
  }
  if (Button == 1){ButtonIsBlocked = false;}
}




/* set the initial Switch State on Power ON
 * needed on future Options configured by the user
 * 
*/
void   setInitialSwitchState() {
    // default should by that Relais is OFF
    if (stateONPowerON == 0){switch_OFF();}
    if (stateONPowerON == 1){switch_ON();}
         
}



//We want some valid values in the EEPROM after the first start 
// So we gonna generate us some
void firstStart() {
  
  if (first_Startup == 0) {
   eeprom_Function = 1;
   Serial.println("#### FIRST STARTUP ####");       
   Serial.println("generating default values now !");       
   Serial.println("Please wait...");
   //We generate the EEPROM-values in a loop and call each Function one after the other
   
   //If you wish to have your own values set to EEPROM then edit and uncomment the first "//" of the following list
   /* Not used right now
   //deepSleepEnabled = 0;                  //Do we want a DeepSleep? 
   //timeDeepSleep = 600;                   //DeepSleep time / sec.
   //waitDeepSleep = 10;                    //wait time before going into DeepSleep / sec.
   //correction_temp = 0.0;                 //correction value to add to temperature measurement 
   //correction_hum = 0.0;                  //correction value to add to humidity measurement 
   //measureTime = 30;                      //time value in sec. to make a sensor measurement 
   //reportDataTime = 120;                  //time value in sec. to submit measurement data to external system
   //CCU_active = 0;                        //do we need to report to CCU ? 0 = inactive 1=active
   //host_CCUIP[16] = "192.168.50.2";       //IP Adress CCU2
   //host_CCUPORT = 80;                     //Portadress CCU2
   //CCU_CuxDAdress[38] = "CUX9002003";     //Name of CCU CuxD Device Adress 
   //CCU_CuxD = 0;                          //do we need to report to CuxD insteadt of CCU ? 0 = inactive 1=active
   //CCU_SV_Temp[80] = "SV_Temp";           //Name of CCU temperature system variable 
   //CCU_SV_Hum[80] = "SV_Hum";             //Name of CCU humidity system variable
   //IOB_active = 0;                        //do we need to report to ioBroker ? 0 = inactive 1=active    
   */
   //host_ioBrokerIP[16] = "0.0.0.0";       //IP Adress ioBroker
   //host_ioBrokerPORT = 8087;              //Portadress simple API Adapter
   //IOB_Datapoint_Path[80] = "esp8266.0";  //Name of ioBroker Datapoint path
   //IOB_Datapoint_Sub[80] = "Path";        //Name of ioBroker Subfolder
   //IOB_Datapoint_State[80] = "State";     //Name of ioBroker State datapoint
   //IOB_Datapoint_Message[80] = "Info";    //Name of ioBroker Info datapoint
   //CCU_CuxDAdress[38] = "CUX2801001:1";   //Name of CCU CuxD Device Adress incl. Channel
   //CCU_CuxD = 0;                          //do we need to report to CuxD ? 0 = inactive 1=active
   //host_CCUIP[16] = "192.0.0.0";          //IP Adress CCU2
   //IOB_active = 0;                        //do we need to report to ioBroker ? 0 = inactive 1=active     

        
   for (int i = 0; i < 13; ++i) {
       //Serial.println("EEPROM_function : " + String(eeprom_Function));       
       writeEEPROM();
       delay(10);
       ++eeprom_Function;
    }
    eeprom_Function = 99;
    first_Startup = 1;
    delay(10);
    writeEEPROM();
    eeprom_Function = 0;
    build_wwwStatus("default values written to EEPROM");
    }

}
void URL_GoBack() {
  wwwGoBack = "<br><a href=\"/\" target=\"_self\">OK - GO BACK</a>";
}

void build_wwwStatus(String S) {

  wwwStatus4 = wwwStatus3; 
  wwwStatus3 = wwwStatus2; 
  wwwStatus2 = wwwStatus1;
  wwwStatus1 = timeStamp() + " " + S;

  wwwStatus = wwwStatus1 + " / " + wwwStatus2 + " / " + wwwStatus3 + " / " + wwwStatus4;
  
}

void createCommandHeader (){
  wwwCommandHeader =  "<!DOCTYPE HTML>\n"
                      "<html><head><meta content=\"text/html;charset=utf-8\"><title>ESP8266 Wall Plug OBI</title>\n"
                      "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\">"                  
                      "<meta http-equiv=\"refresh\" content=\"5; URL=http://" + WiFi.localIP().toString() +"/\">"
                      "<style type=\"text/css\">body {text-align:center;font-family: sans-serif;background-color: #000000;color: #AAAAAA;font-size:0.9em;} a:link, a:visited, a:hover, a:active {color:#fffff;}</style>"
                      "</head>\n"
                      "<body>\n"
                      "<center>\n";
}

